/*

a)


b)


c)


d)


e)


f)


g)


 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"

/**
 * Ett segment av ett spår. För att undvika kollisioner vill vi
 * garantera att det bara finns ett tåg i varje segment.
 */
struct segment {
    // Vilket tåg som finns i segmentet just nu. Noll betyder att
    // segmentet är tomt.
    int occupied_by;
};


/**
 * Ett spår. Varje spår består av ett eller flera segment. I varje
 * segment kan maximalt ett tåg befinna sig.
 */
struct track {
    // Array av segment.
    struct segment *segments;

    // Antal segment i "segments".
    int segment_count;

    // Vilken riktning tågen får röra sig. Antingen 1 eller -1.
    int direction;
};


// Skapa ett spår bestående av ett visst antal segment. "direction" är
// antingen 1 eller -1 och indikerar i vilken riktning tågen får åka.
struct track *track_create(int segment_count, int direction) {
    struct track *track = malloc(sizeof(struct track));
    track->segments = malloc(sizeof(struct segment)*segment_count);
    track->segment_count = segment_count;
    track->direction = direction;

    for (int i = 0; i < segment_count; i++) {
        struct segment *segment = &track->segments[i];
        segment->occupied_by = 0;
    }

    return track;
}

// Frigör minnet som används av ett spår.
void track_free(struct track *track) {
    free(track->segments);
    free(track);
}

// Anropas när ett tåg ska börja åka på spåret. Tågen börjar alltid
// åka i början av spåret (vad som är början beror på "direction").
// Returnerar segmentet som tåget börjar i.
int track_enter(struct track *track, int train_id) {
    int segment_id = 0;
    if (track->direction < 0)
        segment_id = track->segment_count - 1;

    struct segment *segment = &track->segments[segment_id];

    // Vänta på att tåget får plats.
    while (segment->occupied_by != 0)
        ;
    segment->occupied_by = train_id;

    return segment_id;
}

// Anropas när ett tåg ska flytta sig från ett segment till ett
// annat. Tåget flyttar sig från segmentet "from_segment" och ett steg
// i riktningen som anges av spårets riktning.
void track_move(struct track *track, int from_segment) {
    struct segment *from = &track->segments[from_segment];
    struct segment *to = &track->segments[from_segment + track->direction];

    while (to->occupied_by != 0)
        ;
    to->occupied_by = from->occupied_by;
    from->occupied_by = 0;
}

// Anropas när tåget åker av spåret (dvs. när tåget har kommit fram
// till sin destination).
void track_leave(struct track *track, int segment_id) {
    struct segment *segment = &track->segments[segment_id];

    segment->occupied_by = 0;
}

// Rita ut spåret till standard output. Vi antar att ingen annan tråd
// skriver saker till standard output när denna funktion anropas.
void track_draw(struct track *track) {
    // Ovansidan:
    printf("|");
    for (int i = 0; i < track->segment_count; i++) {
        printf("----|");
    }
    printf("\n");

    // Inuti:
    printf("|");
    for (int i = 0; i < track->segment_count; i++) {
        int occupant = track->segments[i].occupied_by;
        if (occupant != 0)
            printf("%4d|", occupant);
        else
            printf("   -|");
    }
    printf("\n");

    printf("|");
    for (int i = 0; i < track->segment_count; i++) {
        printf("----|");
    }
    printf("\n");
}

/**
 * Huvudprogram.
 *
 * Eventuella ändringar här nedanför kommer att ignoreras vid rättning
 * av uppgiften. Du får däremot ändra nedanför ifall du vill testa din
 * lösning.
 */

// Håll koll på vilka nummer som tågen har fått.
int train_id = 1;
struct lock train_id_lock;

// Håll koll på vilka trådar som är klara.
int trains_done = 0;
struct lock done_lock;

// Hjälpfunktion för att sova en slumpmässig tid.
void random_sleep(void) {
    // Notera: "rand" är säker att anropa från flera trådar samtidigt.
    int time = (rand() % 300) + 500;
    timer_msleep(time);
}

// Kör ett tåg längs ett spår. Anropas från main nedan.
void run_train(void *data) {
    struct track *track = data;

    // Hämta ett unikt ID för tåget.
    lock_acquire(&train_id_lock);
    int id = train_id++;
    lock_release(&train_id_lock);

    // Börja från början.
    int position = track_enter(track, id);
    random_sleep();

    // Stega igenom alla segment.
    for (int i = 1; i < track->segment_count; i++) {
        track_move(track, position);
        position += track->direction;

        random_sleep();
    }

    // Sedan är vi klara!
    track_leave(track, position);

    lock_acquire(&done_lock);
    trains_done++;
    lock_release(&done_lock);
}

int main(void) {
    lock_init(&train_id_lock);
    lock_init(&done_lock);

    struct track *outbound = track_create(10, 1);
    struct track *inbound = track_create(10, -1);

    int threads_to_spawn = 6;

    // Starta trådar som representerar tåg i olika riktningar.
    for (int i = 0; i < threads_to_spawn; i++) {
        thread_new(&run_train, outbound);
        thread_new(&run_train, inbound);
    }

    // Uppdatera skärmen tills alla tåg har åkt till sin destination.
    bool done = false;
    while (!done) {
        lock_acquire(&done_lock);
        if (trains_done == 2*threads_to_spawn)
            done = true;
        lock_release(&done_lock);

        // Rensa terminalen.
        printf("\e[1;1H\e[2J");

        // Skriv ut ny data.
        printf("Outbound:\n");
        track_draw(outbound);
        printf("Inbound:\n");
        track_draw(inbound);

        timer_msleep(100);
    }

    track_free(outbound);
    track_free(inbound);

    return 0;
}
